//-------------------------------------------------------------------
// FILENAME    : loadconf.cpp
// DESCRIPTION : Loading Configuration Data
//-------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include "loadconf.h"


const char* Conffile::SPACE_CHARS = " \t";
const char* Conffile::SPACE_AND_EQUAL_CHARS = " \t=";


//-----------------------------------------------------
// constructor, destructor
//-----------------------------------------------------
Conffile::Conffile(const char *fn) : nvars_(0), list_(NULL)
{
	// open file
	FILE *fp = fopen(fn, "rb"); // binary-mode
	if(fp == NULL){
		fprintf(stderr, "can't open %s\nexit\n", fn);
		//		exit(1);
		return ;
	}

	int line;
	char *str;
	for(line = 1; (str = freadline(fp)) != NULL; line++)
	{
		char *p, *q;
		char *varname;
		char *content;

		// remove trailing comments(#)
		p = strpbrk(str, "#;");
		if(p) *p = 0;

		// seek variable name
		varname = seek_word(str);
		if(varname)
		{
			// seek equal(=)
			p = strchr(str, '=');
			if(p)
			{
				// split into varname-part and content-part
				*p = 0;
				content = p + 1;

				// find the end of varname
				for(p = varname; (q = next_word(p)) != NULL; p = q);
				p = strpbrk(p, SPACE_CHARS);
				if(p) *p = 0;

				// seek content
				content = seek_word(content);
				if(content)
				{
					// seek the last word
					for(p = content; (q = next_word(p)) != NULL; p = q);
					p = strpbrk(p, SPACE_CHARS);
					if(p) *p = 0;
				}
				else
					content = ""; // empty

				write(varname, content);
			}
			else
			{
				fprintf(stderr, "LINE %d, no equal(=) found.\n", line);
				destruct();
				free(str);
				fclose(fp);
			}
		}
		free(str);
	}
	fclose(fp);
}


void Conffile::destruct()
{
	int i;
	for(i = 0; i < nvars_; i++)
	{
		if(list_[i].name) free(list_[i].name);
		if(list_[i].value) free(list_[i].value);
	}
	if(list_) free(list_);
}


//-------------------------------------------------
// method
//-------------------------------------------------

int Conffile::write(const char *varname, const char *value)
{
	struct VARINFO *temp;
	int i;

	for(i = 0; i < nvars_; i++)
		if(strcmp(list_[i].name, varname) == 0)
		{
			free(list_[i].value);
			list_[i].value = strdup(value);
			return 0;
		}

	// memory reallocation
	temp = (struct VARINFO *)realloc(
		list_, sizeof(struct VARINFO) * (nvars_ + 1));
	if(temp == NULL)
		return -1;

	list_ = temp;

	// copy strings
	list_[nvars_].name = strdup(varname);
	list_[nvars_].value = strdup(value);
	++nvars_;

	return 0;
}


const char *Conffile::read(const char *varname) const
{
	int i;
	for(i = 0; i < nvars_; i++)
		if(strcmp(list_[i].name, varname) == 0)
			return list_[i].value;

	return NULL;
}




//-----------------------------------------------------
// private static member
//-----------------------------------------------------

// seek the head of the first word
char *Conffile::seek_word(char *p)
{
	p += strspn(p, SPACE_CHARS);
	if(*p == 0)
		return NULL;
	return p;
}


// the next word if available, NULL for nothing to be found
char *Conffile::next_word(char *p)
{
	p = strpbrk(p, SPACE_CHARS);
	if(p == NULL) return NULL;
	p = p + strspn(p, SPACE_CHARS);
	if(*p == 0) return NULL;
	return p;
}


// read a line from file
char *Conffile::freadline(FILE *fp)
{
	const size_t ADDITION = 2000;
	char *buf = NULL;
	size_t size = 0;
	size_t cur = 0;
	int c;

	for(;;)
	{
		if(size <= cur)
		{
			char *temp = (char *)realloc(buf, sizeof(char) * (ADDITION + size));
			if(temp == NULL)
			{
				if(buf) free(buf);
				return NULL;
			}
			buf = temp;
			size += ADDITION;
		}

		c = fgetc(fp);
		if(c == EOF)
		{
			if(cur == 0)
			{
				free(buf);
				return NULL;
			}
			buf[cur] = 0;
			return buf;
		}
		else if(c == '\r')
		{
			c = fgetc(fp);
			if(c != '\n') // for Macintosh format text
				ungetc(c, fp);

			buf[cur] = 0;
			return buf;
		}
		else if(c == '\n')
		{
			buf[cur] = 0;
			return buf;
		}
		else
			buf[cur++] = c;
	}
}



