//-------------------------------------------------------------------
// FILENAME    : one_nsga2_mpi.cpp
// DESCRIPTION : Operating GA  of Parallel NSGA-II
//-------------------------------------------------------------------
#include <stdio.h>
#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <sys/time.h>
#include <mpi.h>
#include "pop.h"
#include "gene.h"
#include "oneisland.h"
#include "one_nsga2.h"
#include "one_nsga2_mpi.h"
#include "base.h"

#define ROOT 0

One_nsga2_mpi::One_nsga2_mpi(int& argc, char** & argv) : One_nsga2()
{

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &procs);
  unsigned long seed = (unsigned long)time(NULL);
  MPI_Bcast(&seed, 1, MPI_INT, 0, MPI_COMM_WORLD);
  mtrand.srand(seed);
  tmpGene=NULL;
  printf("rank %d procs %d start\n",rank,procs);
}

One_nsga2_mpi::~One_nsga2_mpi()
{
  if(procs==1||rank!=ROOT){
    if(tmpGene != NULL){
      delete tmpGene;
      tmpGene = NULL;
    }
  }
  MPI_Finalize();
}

void One_nsga2_mpi::initGA()
{
  if(rank==ROOT)
    One_nsga2::initGA();
  else{
    init_base();
  }
  if(procs==1||rank!=ROOT){
    tmpGene = new Gene(variable.chromosome_length,variable.obj_num);
  }
}

void One_nsga2_mpi::all_open_w_file()
{
  if(rank != ROOT)
    return;
  One_nsga2::all_open_w_file();
  fprintf(fp1,"Let's start Master-Slave of NSGA2 procs= %d\n",procs);
}


void One_nsga2_mpi::cycle()
{
  if(rank == ROOT){
    all_open_w_file();
  }
  int loop_count=0;
  do{
    if(rank==ROOT){
      initialize();
      fprintf(fp1,"loop %d\n",loop_count);
      fprintf(fp3,"loop %d\n",loop_count);
    }
    else{
	set_s_time();
	generation=0;
	func_time=0.0;
	life_flag=1;
	func.re_set_func_count();
	func_count=0;
    }
    if(rank==ROOT)
      f_output();
    terminal_check();
    
     for(;life_flag;){
      if(rank==ROOT)
	printf("####(loop %d)generation %d nsga2 procs %d(pop_num %d)#####\n",loop_count,generation,procs,pop.get_pop_num());
      generation++;
      next_1_gene();
    }
    if(rank==ROOT){
      f_output();
    }
    loop_count++;
  }while(loop_count<variable.loop_num);
  if(rank==ROOT){
    set_all_e_time();
    finish_write();
    all_open_file_close();
  }
}

void One_nsga2_mpi::next_1_gene()
{
  if(rank==ROOT){
    selection();
    cross_over();
    mutation();
  }
  function();
  if(rank==ROOT){
    enviroment_selection();
    f_output();
  }
  terminal_check();
}

void One_nsga2_mpi::terminal_check()
{
  variable_gather();
  if(generation>=variable.end_gene||func.get_func_count() >= variable.limit_func_count){
    life_flag=0;
    set_e_time();
    if(rank==ROOT)
      printf("terminal OK generation%d all_func_count %d \n",generation,func.get_func_count());
  }
  terminal_sign();
}

void One_nsga2_mpi::terminal_sign()
{
  MPI_Bcast(&life_flag,1,MPI_BYTE,ROOT,MPI_COMM_WORLD);
}

void One_nsga2_mpi::variable_gather()
{
  int dum,r_dum;
  if(rank==ROOT){
    dum=0;
  }
  else{
    dum = func.get_func_count();
  }
  MPI_Reduce(&dum,&r_dum,1,MPI_INT,MPI_SUM,ROOT,MPI_COMM_WORLD);
  if(rank==ROOT){
      func.set_func_count(r_dum);
      printf("********variable_gather(): all_func_count %d *************\n",r_dum);
  }
}

int One_nsga2_mpi::get_size_simple(Pop& pop)
{
  int i,size=0;
  for(i=0;i<pop.get_pop_num();i++)
    size += pop.ap_gene[i]->GetRnaLength();
  return size;
}

const u8* One_nsga2_mpi::nslc_transcription(u8 *buffer)
{
  int size=0,i,pop_num;
  pop_num = pop.get_pop_num();
  size =0;
  for(i=0;i<pop_num;i++){
    pop.ap_gene[i]->transcription(buffer+size);
    size += pop.ap_gene[i]->GetRnaLength();
  }
  return buffer;
}

void One_nsga2_mpi::function()
{
  int i,msgt,size,sum_size,recv_num;
  int s_rank;
  MPI_Status stat[2];
  MPI_Request request[2];
  u8 *rna_recv,dum_dum;
  u8 *rna_send;

  rna_recv=NULL;
  if(procs==1){
    size=pop.ap_gene[0]->GetRnaLength();
    rna_send = new u8[size];
    u8 *dum_u;
    dum_u=new u8[size];

    pop.ap_gene[0]->transcription(rna_send);
    tmpGene->reverse_transcription2(rna_send);

    for(i=0;i<pop.get_pop_num();i++)
      func.func(*pop.ap_gene[i]);
    return ;
  }

  if(rank==ROOT){
    MPI_Request f_req[procs-1];
    MPI_Status f_st[procs-1];    
    int count,limit,s_size,dum,gene_size;
    
    gene_size=pop.ap_gene[0]->GetRnaLength();
    s_size = get_size_simple(pop);
    rna_send=new u8[s_size];
    
    nslc_transcription(rna_send); 
    count=dum=0;
    limit = pop.get_pop_num();
    
    msgt=2222;
    sum_size=0;
    for(i=0;i<(procs-1);i++){
      size=pop.ap_gene[i]->GetRnaLength();
      MPI_Isend(rna_send+sum_size,size,MPI_BYTE,i+1,msgt,MPI_COMM_WORLD,&f_req[i]);
      sum_size += size;
    }
    u8 *dum_u;
    dum_u=new u8[size];
    memcpy(dum_u,rna_send,size);
    MPI_Waitall(procs-1,f_req,f_st);
    count=procs-1;

    do{
      msgt=4444;
      MPI_Probe(MPI_ANY_SOURCE,msgt,MPI_COMM_WORLD,&stat[0]);

      s_rank=stat[0].MPI_SOURCE;
      MPI_Get_count(&stat[0],MPI_BYTE,&size);
      rna_recv = new u8[size];
      MPI_Irecv(rna_recv,size,MPI_BYTE,s_rank,msgt,MPI_COMM_WORLD,&request[0]);
      recv_num=size/gene_size;

      if(count<limit){
	msgt = 2222;
	size = pop.ap_gene[count]->GetRnaLength();
	MPI_Isend(rna_send+sum_size,size,MPI_BYTE,s_rank,msgt,MPI_COMM_WORLD,&request[1]);
	sum_size += size;
	MPI_Waitall(2,request,stat);
      }
      else{ 
	MPI_Wait(&request[0],&stat[0]);
      }
      size=0;
      for(i=0;i<recv_num;i++,dum++){
	pop.ap_gene[dum]->reverse_transcription2(rna_recv+size);
	size += pop.ap_gene[dum]->GetRnaLength();
      }
      count++;
      delete [] rna_recv;
    }while(count<(limit+procs-1));
    delete [] rna_send;
    msgt=9999;

    for(i=0;i<(procs-1);i++)
      MPI_Isend(&dum_dum,1,MPI_BYTE,i+1,msgt,MPI_COMM_WORLD,&f_req[i]);
    MPI_Waitall(procs-1,f_req,f_st);
    pop.set_pop_num(dum);
  }
  else{
    int count;
    count=0;
    for(;;){
      MPI_Probe(0,MPI_ANY_TAG,MPI_COMM_WORLD,&stat[0]);
      if(stat[0].MPI_TAG == 9999)
	break;
      msgt=stat[0].MPI_TAG;
      MPI_Get_count(&stat[0],MPI_BYTE,&size);
      try{
	rna_recv = new u8[size];
      }catch(bad_alloc &ba){
	printf("\n**********ERROR *************\n");
      }
      MPI_Irecv(rna_recv,size,MPI_BYTE,ROOT,msgt,MPI_COMM_WORLD,&request[0]);
      MPI_Wait(&request[0],&stat[0]);

      tmpGene->reverse_transcription2(rna_recv);

      func.func(*tmpGene);

      rna_send = new u8[tmpGene->GetRnaLength()];
      tmpGene->transcription(rna_send);
      msgt=4444;
      MPI_Isend(rna_send,size,MPI_BYTE,ROOT,msgt,MPI_COMM_WORLD,&request[0]);
      MPI_Wait(&request[0],&stat[0]);
      count++;
      delete [] rna_recv;
      delete [] rna_send;
    }
    msgt=9999;
    MPI_Recv(&dum_dum,1,MPI_BYTE,ROOT,msgt,MPI_COMM_WORLD,&stat[0]);
  }
}






